/**
 * @author Rafael Ostertag <rafael.ostertag@math.uzh.ch>
 */

/* global $ */

if (!QfqNS) {
    var QfqNS = {};
}

if (!QfqNS.Helper) {
    QfqNS.Helper = {};
}

(function (n) {
    'use strict';

    n.FunctionList = function () {
        this.functions = [];
    };

    n.FunctionList.prototype.addFunction = function (func, thisObj) {
        if (typeof func !== "function") {
            throw new TypeError("Expected function");
        }

        if (thisObj) {
            this.functions.push([func, thisObj]);
        } else {
            this.functions.push(func);
        }
    };

    n.FunctionList.prototype.call = function () {
        var _arguments = arguments;
        this.functions.forEach(function (func) {
            if (typeof func === "function") {
                func.apply(undefined, _arguments);
            } else {
                // func is an array [ Function, thisObj ]
                func[0].apply(func[1], _arguments);
            }
        });
    };
})(QfqNS.Helper);
/**
 * @author Rafael Ostertag <rafael.ostertag@math.uzh.ch>
 */

/* global $ */

if (!QfqNS) {
    var QfqNS = {};
}

(function (n) {
    'use strict';

    /**
     * Display a message.
     *
     * Display one message on a page. Several instances can be used per page, which results in messages being
     * stacked, with the latest message being at the bottom.
     *
     * The first instance displaying a message will append an `alert container` to the body. The last message being
     * dismissed will remove the `alert container`. A typical call sequence might look like:
     *
     *     var alert = new QfqNS.Alert("Text being displayed", "info", "none");
     *     alert.show();
     *
     * Messages may have different background colors (severity levels), controlled by the second argument
     * `messageType` of the constructor. The possible values are
     *
     *  * `"info"`
     *  * `"warning"`
     *  * `"error"`
     *
     * The values are translated into Bootstrap `alert-*` classes internally.
     *
     * Messages can feature clickable buttons, or no buttons at all, in which case a click anywhere on the message
     * will dismiss it. Buttons are controlled by the third argument to the constructor:
     *
     *  * `"okcancel"`
     *  * `"yesno"`
     *  * `"yesnosave"`
     *  * `"none"`
     *
     * Callback functions of the `Ok` or `Yes` button are added by calling Alert#addOkButtonHandler(). Callback
     * functions of the `Cancel` or `No` button are added by calling Alert#addCancelButtonHandler(). Lastly,
     * Alert#addSaveButtonHandler() adds callback functions to the `Save` button.
     *
     * Regardless of the
     *
     *
     * @param message {string} message to be displayed
     * @param messageType {string} type of message, can either be `"info"`, `"warning"`, or `"error"`.
     * @param buttons {string} buttons to be displayed, can either be `"okcancel"`, `"yesno"`, `"yesnosave"`, or `"none"`.
     * When `"none"` is provided, clicking anywhere on the message will dismiss it.
     * @constructor
     */
    n.Alert = function (message, messageType, buttons) {
        this.message = message;
        this.messageType = messageType || "info";
        this.buttons = buttons || "none";

        this.$alertDiv = null;
        this.shown = false;
        // this.timeout < 1 means forever
        this.timeout = 0;
        this.fadeInDuration = 400;
        this.fadeOutDuration = 400;
        this.timerId = null;

        this.userOkButtonHandlers = new QfqNS.Helper.FunctionList();
        this.userCancelButtonHandlers = new QfqNS.Helper.FunctionList();
        this.userSaveButtonHandlers = new QfqNS.Helper.FunctionList();
    };

    /**
     *
     * @private
     */
    n.Alert.prototype.makeAlertContainerSingleton = function () {
        var alertContainer = $("#qfqAlertContainer");
        if (alertContainer.length === 0) {
            // No container so far, create one
            alertContainer = $("<div>").attr("id", "qfqAlertContainer");
            $("body").append(alertContainer);
        }

        return alertContainer;
    };

    /**
     *
     * @returns {number|jQuery}
     * @private
     */
    n.Alert.prototype.countAlertsInAlertContainer = function () {
        return $("#qfqAlertContainer > div").length;
    };

    /**
     * @private
     */
    n.Alert.prototype.removeAlertContainer = function () {
        $("#qfqAlertContainer").remove();
    };

    /**
     * @private
     */
    n.Alert.prototype.getAlertClassBasedOnMessageType = function () {
        switch (this.messageType) {
            case "warning":
                return "alert-warning";
            case "error":
                return "alert-danger";
            /* jshint -W086 */
            default:
                QfqNS.Log.warning("Message type '" + this.messageType + "' unknown. Use default type.");
            case "info":
                return "alert-success";
            /* jshint +W086 */
        }
    };

    /**
     * @private
     */
    n.Alert.prototype.getButtons = function () {
        var buttons = null;
        switch (this.buttons) {
            case 'okcancel':
                buttons = $("<div>")
                    .addClass("alert-buttons")
                    .append(
                    $("<a>")
                        .append("Ok")
                        .addClass("btn btn-default")
                        .click(this.okButtonHandler.bind(this))
                )
                    .append(
                    $("<a>")
                        .append("Cancel")
                        .addClass("btn btn-default")
                        .click(this.cancelButtonHandler.bind(this))
                );
                return buttons;
            case "yesno":
                buttons = $("<div>")
                    .addClass("alert-buttons")
                    .append(
                    $("<a>")
                        .append("Yes")
                        .addClass("btn btn-default")
                        .click(this.okButtonHandler.bind(this))
                )
                    .append(
                    $("<a>")
                        .append("No")
                        .addClass("btn btn-default")
                        .click(this.cancelButtonHandler.bind(this))
                );
                return buttons;
            case "yesnosave":
                buttons = $("<div>")
                    .addClass("alert-buttons")
                    .append(
                    $("<a>")
                        .append("Yes")
                        .addClass("btn btn-default")
                        .click(this.okButtonHandler.bind(this))
                )
                    .append(
                    $("<a>")
                        .append("No")
                        .addClass("btn btn-default")
                        .click(this.cancelButtonHandler.bind(this))
                )
                    .append(
                    $("<a>")
                        .append("Save & Close")
                        .addClass("btn btn-default")
                        .click(this.saveButtonHandler.bind(this))
                )
                ;
                return buttons;
            /* jshint -W086 */
            default:
                QfqNS.Log.warning("Buttons '" + this.buttons + "' unknown. Use default buttons");
            case "none":
                break;
            /* jshint +W086 */
        }

        return buttons;
    };

    n.Alert.prototype.show = function () {
        if (this.shown) {
            // We only allow showing once
            return;
        }

        var $alertContainer = this.makeAlertContainerSingleton();
        this.$alertDiv = $("<div>")
            .addClass("alert")
            .addClass(this.getAlertClassBasedOnMessageType())
            .attr("role", "alert")
            .append(this.message);


        var buttons = this.getButtons();
        if (buttons && this.timeout < 1) {
            // Buttons will take care of removing the message
            this.$alertDiv.append(buttons);
        } else {
            // Click on the message anywhere will remove the message
            this.$alertDiv.click(this.removeAlert.bind(this));
        }

        $alertContainer.append(this.$alertDiv);
        this.$alertDiv.fadeIn(this.fadeInDuration, this.afterFadeIn.bind(this));
        this.shown = true;


    };

    /**
     * @private
     */
    n.Alert.prototype.afterFadeIn = function () {
        if (this.timeout > 0) {
            this.timerId = window.setTimeout(this.removeAlert.bind(this), this.timeout);
        }
    };

    /**
     *
     * @param event
     *
     * @private
     */
    n.Alert.prototype.removeAlert = function (event) {
        if (!event || event.type !== "click") {
            // No user click, so it must be a timer event
            if (this.timerId) {
                window.clearTimeout(this.timerId);
                this.timerId = null;
            } else {
                QfqNS.Log.error("Alert.remove(): Identified timer event, but had no timer id");
            }
        }
        var that = this;
        this.$alertDiv.fadeOut(this.fadeOutDuration, function () {
            that.$alertDiv.remove();
            that.$alertDiv = null;
            that.shown = false;

            if (that.countAlertsInAlertContainer() === 0) {
                that.removeAlertContainer();
            }
        });

    };

    n.Alert.prototype.addOkButtonHandler = function (handler) {
        this.userOkButtonHandlers.addFunction(handler);
    };

    n.Alert.prototype.addCancelButtonHandler = function (handler) {
        this.userCancelButtonHandlers.addFunction(handler);
    };

    n.Alert.prototype.addSaveButtonHandler = function (handler) {
        this.userSaveButtonHandlers.addFunction(handler);
    };


    /**
     *
     * @param handler
     *
     * @private
     */
    n.Alert.prototype.okButtonHandler = function (handler) {
        this.removeAlert();
        this.userOkButtonHandlers.call(this);
    };

    /**
     *
     * @param handler
     *
     * @private
     */
    n.Alert.prototype.saveButtonHandler = function (handler) {
        this.removeAlert();
        this.userSaveButtonHandlers.call(this);
    };

    /**
     *
     * @param handler
     *
     * @private
     */
    n.Alert.prototype.cancelButtonHandler = function (handler) {
        this.removeAlert();
        this.userCancelButtonHandlers.call(this);
    };

    n.Alert.prototype.isShown = function () {
        return this.shown;
    };

})(QfqNS);
/**
 * @author Rafael Ostertag <rafael.ostertag@math.uzh.ch>
 */

/* global $ */
/* global console */

if (!QfqNS) {
    var QfqNS = {};
}

(function (n) {
    'use strict';
    /**
     * Tab Constructor.
     *
     * @param {string} tabId HTML id of the element having `nav` and `nav-tabs` classes
     * @constructor
     */
    n.BSTabs = function (tabId) {
        this.tabId = tabId;
        this._tabContainerLinkSelector = '#' + this.tabId + ' a[data-toggle="tab"]';
        this._tabActiveSelector = '#' + this.tabId + ' .active a[data-toggle="tab"]';
        this.tabs = {};
        this.currentTab = this.getActiveTabFromDOM();
        this.userTabShowHandlers = new QfqNS.Helper.FunctionList();


        // Fill this.tabs
        this.fillTabInformation();

        // Enable update of current tab field
        this.installTabHandlers();
    };

    /**
     * Get active tab from DOM
     *
     * @private
     */
    n.BSTabs.prototype.getActiveTabFromDOM = function () {
        var activeTabAnchors = $(this._tabActiveSelector);
        if (activeTabAnchors.length < 1) {
            // This could be due to the DOM not fully loaded. If that's really the case, then the active tab
            // attribute should be set by the show.bs.tab handler
            return null;
        }

        return activeTabAnchors[0].hash.slice(1);
    };

    /**
     * Fill tabs object.
     *
     * Fill the tabs object using the tab HTML id as attribute name
     *
     * @private
     */
    n.BSTabs.prototype.fillTabInformation = function () {
        var tabLinks = $(this._tabContainerLinkSelector);
        if ($(tabLinks).length === 0) {
            throw new Error("Unable to find a BootStrap container matching: " + this._tabContainerLinkSelector);
        }

        var that = this;
        tabLinks.each(function (index, element) {
                if (element.hash !== "") {
                    var tabId = element.hash.slice(1);

                    that.tabs[tabId] = {
                        index: index,
                        element: element
                    };
                }
            }
        );
    };

    /**
     * @private
     */
    n.BSTabs.prototype.installTabHandlers = function () {
        $(this._tabContainerLinkSelector)
            .on('show.bs.tab', this.tabShowHandler.bind(this));

    };

    /**
     * Tab Show handler.
     *
     * Sets this.currentTab to the clicked tab and calls all registered tab click handlers.
     *
     * @private
     * @param event
     */
    n.BSTabs.prototype.tabShowHandler = function (event) {
        QfqNS.Log.debug('Enter: BSTabs.tabShowHandler()');
        this.currentTab = event.target.hash.slice(1);

        var that = this;
        QfqNS.Log.debug("BSTabs.tabShowHandler(): invoke user handler(s)");
        this.userTabShowHandlers.call(that);
        QfqNS.Log.debug('Exit: BSTabs.tabShowHandler()');
    };

    /**
     * Get all tab IDs.
     *
     * @returns {Array}
     *
     * @public
     */
    n.BSTabs.prototype.getTabIds = function () {
        var tabIds = [];
        for (var tabId in this.tabs) {
            if (this.tabs.hasOwnProperty(tabId)) {
                tabIds.push(tabId);
            }
        }
        return tabIds;
    };

    /**
     *
     * @returns {Array}
     *
     * @public
     */
    n.BSTabs.prototype.getTabAnchors = function () {
        var tabLinks = [];
        for (var tabId in this.tabs) {
            if (this.tabs.hasOwnProperty(tabId)) {
                tabLinks.push(this.tabs[tabId].element);
            }
        }

        return tabLinks;
    };

    /**
     * Activate a given tab.
     *
     * @param {string} tabId Id of the tab to activate
     *
     */
    n.BSTabs.prototype.activateTab = function (tabId) {
        if (!this.tabs[tabId]) {
            console.error("Unable to find tab with id: " + tabId);
            return false;
        }

        $(this.tabs[tabId].element).tab('show');
        return true;
    };

    n.BSTabs.prototype.getCurrentTab = function () {
        return this.currentTab;
    };

    /**
     * Add tab show handler.
     *
     * @param {function} handler handler function. `this` will be passed as first and only argument to the handler.
     */
    n.BSTabs.prototype.addTabShowHandler = function (handler) {
        this.userTabShowHandlers.addFunction(handler);
    };

    n.BSTabs.prototype.getTabName = function (tabId) {
        if (!this.tabs[tabId]) {
            console.error("Unable to find tab with id: " + tabId);
            return null;
        }

        return $(this.tabs[tabId].element).text().trim();
    };

    n.BSTabs.prototype.getActiveTab = function () {
        return this.currentTab;
    };

    n.BSTabs.prototype.getContainingTabIdForFormControl = function (formControlName) {
        var $formControl = $("[name=" + formControlName + "]");
        if ($formControl.length === 0) {
            QfqNS.Log.debug("BSTabs.getContainingTabForFormControl(): unable to find form control with name '" + formControlName + "'");
            return null;
        }

        var iterator = $formControl[0];
        while (iterator !== null) {
            if (iterator.hasAttribute('role') &&
                iterator.getAttribute('role') === 'tabpanel') {
                return iterator.id || null;
            }

            iterator = iterator.parentElement;
        }

        return null;
    };

})(QfqNS);
/**
 * @author Rafael Ostertag <rafael.ostertag@math.uzh.ch>
 */

/* global $ */

if (!QfqNS) {
    var QfqNS = {};
}

(function (n) {
    'use strict';

    n.Form = function (formId) {
        this.formId = formId;

        if (!document.forms[this.formId]) {
            throw new Error("Form '" + formId + "' does not exist.");
        }

        this.formChanged = false;
        this.userFormChangeHandlers = new QfqNS.Helper.FunctionList();
        this.userResetHandlers = new QfqNS.Helper.FunctionList();
        this.userSubmitSuccessHandlers = new QfqNS.Helper.FunctionList();
        this.userSubmitFailureHandlers = new QfqNS.Helper.FunctionList();

        this.$form = $(document.forms[this.formId]);
        this.$form.on("change", this.changeHandler.bind(this));

        // On text input elements, we specifically bind this events, in order to update the formChanged property
        // immediately, not only after loosing focus.
        this.$form.find("input[type=text]").on("input paste", this.changeHandler.bind(this));
    };

    /**
     *
     * @param event
     *
     * @private
     */
    n.Form.prototype.changeHandler = function (event) {
        this.formChanged = true;
        this.userFormChangeHandlers.call(this);
    };

    n.Form.prototype.addChangeHandler = function (callback) {
        this.userFormChangeHandlers.addFunction(callback);
    };

    n.Form.prototype.addResetHandler = function (callback) {
        this.userResetHandlers.addFunction(callback);
    };

    n.Form.prototype.addSubmitSuccessHandler = function (callback) {
        this.userSubmitSuccessHandlers.addFunction(callback);
    };

    n.Form.prototype.addSubmitFailureHandler = function (callback) {
        this.userSubmitFailureHandlers.addFunction(callback);
    };

    n.Form.prototype.getFormChanged = function () {
        return this.formChanged;
    };

    n.Form.prototype.resetFormChanged = function () {
        this.formChanged = false;
        this.userResetHandlers.call(this);
    };

    n.Form.prototype.submitTo = function (to) {
        $.post(to, this.$form.serialize())
            .done(this.ajaxSuccessHandler.bind(this))
            .fail(this.submitFailureHandler.bind(this));
    };

    /**
     *
     * @param data
     * @param textStatus
     * @param jqXHR
     *
     * @private
     */
    n.Form.prototype.ajaxSuccessHandler = function (data, textStatus, jqXHR) {
        this.userSubmitSuccessHandlers.call(this, data, textStatus);
    };

    /**
     *
     *
     * @private
     */
    n.Form.prototype.submitFailureHandler = function (jqXHR, textStatus, errorThrown) {
        this.userSubmitFailureHandlers.call(this, textStatus, jqXHR, errorThrown);
    };

})(QfqNS);

/**
 * @author Rafael Ostertag <rafael.ostertag@math.uzh.ch>
 */

/* global console */

if (!QfqNS) {
    var QfqNS = {};
}

(function (n) {
    'use strict';
    n.Log = {
        level: 3,
        message: function (msg) {
            if (this.level <= 0) {
                console.log('[message] ' + msg);
            }
        },
        debug: function (msg) {
            if (this.level <= 1) {
                console.log('[debug] ' + msg);
            }
        },
        warning: function (msg) {
            if (this.level <= 2) {
                console.log('[warning] ' + msg);
            }
        },
        error: function (msg) {
            if (this.level <= 3) {
                console.log('[error] ' + msg);
            }
        }

    };
})(QfqNS);
/**
 * @author Rafael Ostertag <rafael.ostertag@math.uzh.ch>
 */

if (!QfqNS) {
    var QfqNS = {};
}

(function (n) {
    'use strict';


    n.PageState = function () {
        this.pageState = location.hash.slice(1);
        this.data = null;
        this.inPoppingHandler = false;
        this.userPopStateHandlers = new QfqNS.Helper.FunctionList();

        window.addEventListener("popstate", this.popStateHandler.bind(this));
    };
    /**
     *
     * @param event
     *
     * @private
     */
    n.PageState.prototype.popStateHandler = function (event) {

        QfqNS.Log.debug("Enter: PageState.popStateHandler()");

        this.inPoppingHandler = true;
        this.pageState = location.hash.slice(1);
        this.data = window.history.state;

        QfqNS.Log.debug("PageState.popStateHandler(): invoke user pop state handler(s)");

        this.userPopStateHandlers.call(this);

        this.inPoppingHandler = false;
        QfqNS.Log.debug("Exit: PageState.popStateHandler()");

    };

    n.PageState.prototype.getPageState = function () {
        return this.pageState;
    };

    n.PageState.prototype.getPageData = function () {
        return this.data;
    };

    n.PageState.prototype.setPageState = function (state, data) {
        if (state.startsWith('#')) {
            this.pageState = state.slice(1);
            window.history.replaceState(data, null, state);
        } else {
            this.pageState = state;
            window.history.replaceState(data, null, '#' + state);
        }
        this.data = data;
    };

    n.PageState.prototype.newPageState = function (state, data) {
        if (state.startsWith('#')) {
            this.pageState = state.slice(1);
            window.history.pushState(data, null, state);
        } else {
            this.pageState = state;
            window.history.pushState(data, null, '#' + state);
        }

        this.data = data;
    };

    n.PageState.prototype.addStateActivationHandler = function (handler) {
        this.userPopStateHandlers.addFunction(handler);
    };


})(QfqNS);
/**
 * @author Rafael Ostertag <rafael.ostertag@math.uzh.ch>
 */


if (!QfqNS) {
    var QfqNS = {};
}

(function (n) {
    'use strict';

    n.PageTitle = {
        set: function (title) {
            document.title = title;
        },
        get: function () {
            return document.title;
        },
        setSubTitle: function (subTitle) {
            var currentTitle = this.get();
            var subtitleStrippedOff = currentTitle.replace(/ - (.*)$/, '');
            document.title = subtitleStrippedOff + " - (" + subTitle + ")";
        }
    };
})(QfqNS);
/**
 * @author Rafael Ostertag <rafael.ostertag@math.uzh.ch>
 */

/* global $ */

if (!QfqNS) {
    var QfqNS = {};
}

(function (n) {
    'use strict';

    n.QfqForm = function (formId, submitTo, deleteUrl) {
        this.formId = formId;
        this.submitTo = submitTo;
        this.deleteUrl = deleteUrl;
        this.form = new n.Form(this.formId);
        this.bsTabs = null;
        this.lastButtonPress = null;
        this.destroyFormUserCallbacks = new n.Helper.FunctionList();

        this.getSaveButton().addClass("disabled").attr("disabled", "disabled");

        this.form.addChangeHandler(this.changeHandler.bind(this));
        this.form.addResetHandler(this.resetHandler.bind(this));
        this.form.addSubmitSuccessHandler(this.submitSuccessDispatcher.bind(this));
        this.form.addSubmitFailureHandler(this.submitFailureHandler.bind(this));

        this.getSaveButton().click(this.handleSaveClick.bind(this));
        this.getCloseButton().click(this.handleCloseClick.bind(this));
        this.getNewButton().click(this.handleNewClick.bind(this));
        this.getDeleteButton().click(this.handleDeleteClick.bind(this));
    };

    n.QfqForm.prototype.setBsTabs = function (bsTabs) {
        this.bsTabs = bsTabs;
    };

    /**
     *
     * @param $button
     * @param enabled {boolean}
     *
     * @private
     */
    n.QfqForm.prototype.setButtonEnabled = function ($button, enabled) {
        if (!$button) {
            n.Log.error("QfqForm#setButtonEnabled(): no button provided.");
            return;
        }
        if (!enabled) {
            $button.addClass("disabled");
            $button.attr("disabled", "disabled");
        } else {
            $button.removeClass("disabled");
            $button.removeAttr("disabled");
        }
    };

    /**
     * @private
     */
    n.QfqForm.prototype.destroyFormAndSetText = function (text) {
        this.form = null;
        $('#' + this.formId).replaceWith($("<p>").append(text));
        this.destroyFormUserCallbacks.call();
    };

    /**
     * @private
     */
    n.QfqForm.prototype.handleSaveClick = function () {
        this.lastButtonPress = "save";
        QfqNS.Log.debug("save click");
        // First, remove all validation states, in case a previous submit has set a validation state, thus we're not
        // stockpiling them.
        this.clearAllValidationStates();
        this.form.submitTo(this.submitTo);
    };

    /**
     * @private
     */
    n.QfqForm.prototype.handleCloseClick = function () {
        this.lastButtonPress = "close";
        if (this.form.getFormChanged()) {
            var alert = new QfqNS.Alert("You have unsaved changes. Do you want to close?", "warning", "yesnosave");
            var that = this;
            alert.addSaveButtonHandler(function () {
                that.form.submitTo(that.submitTo);
            });
            alert.addOkButtonHandler(function () {
                window.history.back();
            });
            alert.show();
        } else {
            window.history.back();
        }
    };

    /**
     * @private
     */
    n.QfqForm.prototype.handleNewClick = function () {
        this.lastButtonPress = "new";
        QfqNS.Log.debug("new click");
    };

    /**
     * @private
     */
    n.QfqForm.prototype.handleDeleteClick = function () {
        this.lastButtonPress = "delete";
        QfqNS.Log.debug("delete click");
        var alert = new QfqNS.Alert("Do you really want to delete the record?", "warning", "yesno");
        var that = this;
        alert.addOkButtonHandler(function () {
            $.post(that.deleteUrl)
                .done(that.ajaxDeleteSuccessDispatcher.bind(that))
                .fail(that.ajaxDeleteFailureHandler.bind(that));
        });
        alert.show();
    };

    /**
     *
     * @param data
     * @param textStatus
     * @param jqXHR
     *
     * @private
     */
    n.QfqForm.prototype.ajaxDeleteSuccessDispatcher = function (data, textStatus, jqXHR) {
        if (!data.status) {
            throw new Error("No 'status' property 'data'");
        }

        switch (data.status) {
            case "error":
                this.handleLogicDeleteError(data);
                break;
            case "success":
                this.handleDeleteSuccess(data);
                break;
            default:
                throw new Error("Status '" + data.status + "' unknown.");
        }
    };

    /**
     *
     * @param data
     *
     * @private
     */
    n.QfqForm.prototype.handleDeleteSuccess = function (data) {
        this.setButtonEnabled(this.getCloseButton(), false);
        this.setButtonEnabled(this.getDeleteButton(), false);
        this.setButtonEnabled(this.getSaveButton(), false);
        this.setButtonEnabled(this.getNewButton(), false);
        this.destroyFormAndSetText("Record has been deleted!");

        if (!data.redirect || data.redirect === "client") {
            window.history.back();
            return;
        }

        if (data.redirect === "no") {
            var alert = new QfqNS.Alert("redirect=='no' not allowed", "error");
            alert.show();
            return;
        }

        if (data.redirect === "url" || data['redirect-url']) {
            window.location = data['redirect-url'];
            return;
        }
    };

    /**
     *
     * @param data
     *
     * @private
     */
    n.QfqForm.prototype.handleLogicDeleteError = function (data) {
        if (!data.message) {
            throw Error("Status is 'error' but required 'message' attribute is missing.");
        }
        var alert = new QfqNS.Alert(data.message, "error");
        alert.show();
    };

    /**
     *
     * @param form {QfqNS.QfqForm}
     *
     * @private
     */
    n.QfqForm.prototype.changeHandler = function (form) {
        this.getSaveButton().removeClass("disabled");
        this.getSaveButton().addClass("alert-warning");
        this.getSaveButton().removeAttr("disabled");
    };

    /**
     *
     * @param form {QfqNS.QfqForm}
     *
     * @private
     */
    n.QfqForm.prototype.resetHandler = function (form) {
        this.getSaveButton().removeClass("alert-warning");
        this.getSaveButton().addClass("disabled");
        this.getSaveButton().attr("disabled", "disabled");
    };

    /**
     *
     * @returns {jQuery|HTMLElement}
     *
     * @private
     */
    n.QfqForm.prototype.getSaveButton = function () {
        return $("#save-button");
    };

    /**
     *
     * @returns {jQuery|HTMLElement}
     *
     * @private
     */
    n.QfqForm.prototype.getCloseButton = function () {
        return $("#close-button");
    };

    /**
     *
     * @returns {jQuery|HTMLElement}
     *
     * @private
     */
    n.QfqForm.prototype.getDeleteButton = function () {
        return $("#delete-button");
    };

    /**
     *
     * @returns {jQuery|HTMLElement}
     *
     * @private
     */
    n.QfqForm.prototype.getNewButton = function () {
        return $("#new-button");
    };

    /**
     *
     * @param jqXHR
     * @param textStatus
     * @param errorThrown
     *
     * @private
     */
    n.QfqForm.prototype.submitFailureHandler = function (form, textStatus, jqXHR, errorThrown) {
        this.showAjaxError(errorThrown);
    };

    /**
     *
     * @param jqXHR
     * @param textStatus
     * @param errorThrown
     *
     * @private
     */
    n.QfqForm.prototype.ajaxDeleteFailureHandler = function (jqXHR, textStatus, errorThrown) {
        this.showAjaxError(errorThrown);
    };

    n.QfqForm.prototype.showAjaxError = function (errorThrown) {
        var alert = new QfqNS.Alert("Error:<br> " +
            errorThrown, "error");
        alert.show();
    };


    /**
     * @private
     */
    n.QfqForm.prototype.submitSuccessDispatcher = function (form, data, textStatus) {
        if (!data.status) {
            throw new Error("No 'status' property in 'data'");
        }

        switch (data.status) {
            case "error":
                this.handleLogicSubmitError(form, data);
                break;
            case "success":
                this.handleSubmitSuccess(form, data);
                break;
            default:
                throw new Error("Status '" + data.status + "' unknown.");
        }

    };

    /**
     *
     * @param form
     * @param data
     *
     * @private
     */
    n.QfqForm.prototype.handleLogicSubmitError = function (form, data) {
        if (!data.message) {
            throw Error("Status is 'error' but required 'message' attribute is missing.");
        }
        var alert = new QfqNS.Alert(data.message, "error");
        alert.show();

        if (data["field-name"] && this.bsTabs) {
            var tabId = this.bsTabs.getContainingTabIdForFormControl(data["field-name"]);
            if (tabId) {
                this.bsTabs.activateTab(tabId);
            }

            this.setValidationState(data["field-name"], "error");
            this.setHelpBlockValidationMessage(data["field-name"], data["field-message"]);
        }
    };

    /**
     *
     * @param form
     * @param data
     *
     * @private
     */
    n.QfqForm.prototype.handleSubmitSuccess = function (form, data) {
        QfqNS.Log.debug('Reset form state');
        form.resetFormChanged();

        if (this.lastButtonPress === 'save' || !data.redirect || data.redirect === "no") {
            if (data.message) {
                var alert = new QfqNS.Alert(data.message);
                alert.timeout = 1500;
                alert.show();
            }
            return;
        }

        if (data.redirect === "client") {
            window.history.back();
            return;
        }


        if (data.redirect === "url" || data['redirect-url']) {
            window.location = data['redirect-url'];
            return;
        }

    };

    n.QfqForm.prototype.getFormGroupByControlName = function (formControlName) {
        var $formControl = $("[name=" + formControlName + "]");
        if ($formControl.length === 0) {
            QfqNS.Log.debug("QfqForm.setValidationState(): unable to find form control with name '" + formControlName + "'");
            return null;
        }

        var iterator = $formControl[0];
        while (iterator !== null) {
            var $iterator = $(iterator);
            if ($iterator.hasClass('form-group')) {
                return $iterator;
            }

            iterator = iterator.parentElement;
        }

        return null;
    };

    n.QfqForm.prototype.setValidationState = function (formControlName, state) {
        var $formGroup = this.getFormGroupByControlName(formControlName);
        if ($formGroup) {
            $formGroup.addClass("has-" + state);
        }
    };

    n.QfqForm.prototype.resetValidationState = function (formControlName) {
        var $formGroup = this.getFormGroupByControlName(formControlName);
        $formGroup.removeClass("has-warning");
        $formGroup.removeClass("has-error");
        $formGroup.removeClass("has-success");
    };

    n.QfqForm.prototype.clearAllValidationStates = function () {
        $('.has-warning,.has-error,.has-success').removeClass("has-warning has-error has-success");
        $('[data-qfq=validation-message]').remove();
    };

    n.QfqForm.prototype.setHelpBlockValidationMessage = function (formControlName, text) {
        var $formGroup = this.getFormGroupByControlName(formControlName);
        if (!$formGroup) {
            return;
        }
        var $helpBlockColumn;
        var $formGroupSubDivs = $formGroup.find("div");
        if ($formGroupSubDivs.length < 3) {
            $helpBlockColumn = $("<div>").addClass("col-md-4");
            $formGroup.append($helpBlockColumn);
        } else {
            $helpBlockColumn = $($formGroupSubDivs[2]);
        }

        $helpBlockColumn.append(
            $("<p>")
                .addClass("help-block")
                .attr("data-qfq", "validation-message")
                .append(text)
        );
    };

})(QfqNS);
/**
 * @author Rafael Ostertag <rafael.ostertag@math.uzh.ch>
 */

/* global $ */
/* global console */

if (!QfqNS) {
    var QfqNS = {};
}

(function (n) {
    'use strict';

    n.QfqPage = function (settings) {
        this.settings = $.extend(
            {
                tabsId: "qfqTabs",
                formId: "qfqForm",
                submitTo: "typo3conf/ext/qfq/qfq/api/save.php",
                deleteUrl: "typo3conf/ext/qfq/qfq/api/delete.php",
                pageState: new QfqNS.PageState()
            }, settings
        );

        try {
            this.bsTabs = new QfqNS.BSTabs(this.settings.tabsId);

            var currentState = this.settings.pageState.getPageState();
            if (currentState !== "") {
                this.bsTabs.activateTab(currentState);
                QfqNS.PageTitle.setSubTitle(this.bsTabs.getTabName(currentState));
            } else {
                this.settings.pageState.setPageState(this.bsTabs.getCurrentTab(), QfqNS.PageTitle.get());
            }


            this.bsTabs.addTabShowHandler(this.tabShowHandler.bind(this));
            this.settings.pageState.addStateActivationHandler(this.popStateHandler.bind(this));
        } catch (e) {
            QfqNS.Log.message(e.message);
            this.bsTabs = null;
        }

        try {
            this.qfqForm = new QfqNS.QfqForm(this.settings.formId, this.settings.submitTo, this.settings.deleteUrl);
            this.qfqForm.setBsTabs(this.bsTabs);
            this.qfqForm.destroyFormUserCallbacks.addFunction(this.destroyFormHandler.bind(this));
        } catch (e) {
            QfqNS.Log.error(e.message);
            this.qfqForm = null;
        }
    };

    /**
     * @private
     */
    n.QfqPage.prototype.destroyFormHandler = function () {
        this.settings.qfqForm = null;
        $('#' + this.settings.tabsId).remove();
    };

    n.QfqPage.prototype.tabShowHandler = function (bsTabs) {
        // tabShowHandler will be called every time the tab will be shown, regardless of whether or not this happens
        // because of BSTabs.activateTab() or user interaction.
        //
        // Therefore, we have to make sure, that tabShowHandler() does not save the page state while we're restoring
        // a previous state, i.e. we're called because of the popStateHandler() below.
        if (this.settings.pageState.inPoppingHandler) {
            QfqNS.Log.debug("Prematurely terminating QfqPage.tabShowHandler(): called due to page state" +
                " restoration.");
            return;
        }
        var currentTabId = bsTabs.getCurrentTab();
        QfqNS.Log.debug('Saving state: ' + currentTabId);
        QfqNS.PageTitle.setSubTitle(bsTabs.getTabName(currentTabId));
        this.settings.pageState.setPageState(currentTabId, QfqNS.PageTitle.get());

    };

    n.QfqPage.prototype.popStateHandler = function (pageState) {
        this.bsTabs.activateTab(pageState.getPageState());
        QfqNS.PageTitle.set(pageState.getPageData());
    };

})(QfqNS);