/**
 * @author Rafael Ostertag <rafael.ostertag@math.uzh.ch>
 */

/* global $ */

if (!QfqNS) {
    var QfqNS = {};
}

(function (n) {
    'use strict';

    /**
     * Display a message.
     *
     * Display one message on a page. Several instances can be used per page, which results in messages being
     * stacked, with the latest message being at the bottom.
     *
     * The first instance displaying a message will append an `alert container` to the body. The last message being
     * dismissed will remove the `alert container`. A typical call sequence might look like:
     *
     *     var alert = new QfqNS.Alert("Text being displayed", "info", "none");
     *     alert.show();
     *
     * Messages may have different background colors (severity levels), controlled by the second argument
     * `messageType` of the constructor. The possible values are
     *
     *  * `"info"`
     *  * `"warning"`
     *  * `"error"`
     *
     * The values are translated into Bootstrap `alert-*` classes internally.
     *
     * Messages can feature clickable buttons, or no buttons at all, in which case a click anywhere on the message
     * will dismiss it. Buttons are controlled by the third argument to the constructor:
     *
     *  * `"okcancel"`
     *  * `"yesno"`
     *  * `"yesnosave"`
     *  * `"none"`
     *
     * Callback functions of the `Ok` or `Yes` button are added by calling Alert#addOkButtonHandler(). Callback
     * functions of the `Cancel` or `No` button are added by calling Alert#addCancelButtonHandler(). Lastly,
     * Alert#addSaveButtonHandler() adds callback functions to the `Save` button.
     *
     * Regardless of the
     *
     *
     * @param message {string} message to be displayed
     * @param messageType {string} type of message, can either be `"info"`, `"warning"`, or `"error"`.
     * @param buttons {string} buttons to be displayed, can either be `"okcancel"`, `"yesno"`, `"yesnosave"`, or `"none"`.
     * When `"none"` is provided, clicking anywhere on the message will dismiss it.
     * @constructor
     */
    n.Alert = function (message, messageType, buttons) {
        this.message = message;
        this.messageType = messageType || "info";
        this.buttons = buttons || "none";

        this.$alertDiv = null;
        this.shown = false;
        // this.timeout < 1 means forever
        this.timeout = 0;
        this.fadeInDuration = 400;
        this.fadeOutDuration = 400;
        this.timerId = null;

        this.userOkButtonHandlers = new QfqNS.Helper.FunctionList();
        this.userCancelButtonHandlers = new QfqNS.Helper.FunctionList();
        this.userSaveButtonHandlers = new QfqNS.Helper.FunctionList();
    };

    /**
     *
     * @private
     */
    n.Alert.prototype.makeAlertContainerSingleton = function () {
        var alertContainer = $("#qfqAlertContainer");
        if (alertContainer.length === 0) {
            // No container so far, create one
            alertContainer = $("<div>").attr("id", "qfqAlertContainer");
            $("body").append(alertContainer);
        }

        return alertContainer;
    };

    /**
     *
     * @returns {number|jQuery}
     * @private
     */
    n.Alert.prototype.countAlertsInAlertContainer = function () {
        return $("#qfqAlertContainer > div").length;
    };

    /**
     * @private
     */
    n.Alert.prototype.removeAlertContainer = function () {
        $("#qfqAlertContainer").remove();
    };

    /**
     * @private
     */
    n.Alert.prototype.getAlertClassBasedOnMessageType = function () {
        switch (this.messageType) {
            case "warning":
                return "alert-warning";
            case "error":
                return "alert-danger";
            /* jshint -W086 */
            default:
                QfqNS.Log.warning("Message type '" + this.messageType + "' unknown. Use default type.");
            case "info":
                return "alert-success";
            /* jshint +W086 */
        }
    };

    /**
     * @private
     */
    n.Alert.prototype.getButtons = function () {
        var buttons = null;
        switch (this.buttons) {
            case 'okcancel':
                buttons = $("<div>")
                    .addClass("alert-buttons")
                    .append(
                    $("<a>")
                        .append("Ok")
                        .addClass("btn btn-default")
                        .click(this.okButtonHandler.bind(this))
                )
                    .append(
                    $("<a>")
                        .append("Cancel")
                        .addClass("btn btn-default")
                        .click(this.cancelButtonHandler.bind(this))
                );
                return buttons;
            case "yesno":
                buttons = $("<div>")
                    .addClass("alert-buttons")
                    .append(
                    $("<a>")
                        .append("Yes")
                        .addClass("btn btn-default")
                        .click(this.okButtonHandler.bind(this))
                )
                    .append(
                    $("<a>")
                        .append("No")
                        .addClass("btn btn-default")
                        .click(this.cancelButtonHandler.bind(this))
                );
                return buttons;
            case "yesnosave":
                buttons = $("<div>")
                    .addClass("alert-buttons")
                    .append(
                    $("<a>")
                        .append("Yes")
                        .addClass("btn btn-default")
                        .click(this.okButtonHandler.bind(this))
                )
                    .append(
                    $("<a>")
                        .append("No")
                        .addClass("btn btn-default")
                        .click(this.cancelButtonHandler.bind(this))
                )
                    .append(
                    $("<a>")
                        .append("Save & Close")
                        .addClass("btn btn-default")
                        .click(this.saveButtonHandler.bind(this))
                )
                ;
                return buttons;
            /* jshint -W086 */
            default:
                QfqNS.Log.warning("Buttons '" + this.buttons + "' unknown. Use default buttons");
            case "none":
                break;
            /* jshint +W086 */
        }

        return buttons;
    };

    n.Alert.prototype.show = function () {
        if (this.shown) {
            // We only allow showing once
            return;
        }

        var $alertContainer = this.makeAlertContainerSingleton();
        this.$alertDiv = $("<div>")
            .addClass("alert")
            .addClass(this.getAlertClassBasedOnMessageType())
            .attr("role", "alert")
            .append(this.message);


        var buttons = this.getButtons();
        if (buttons && this.timeout < 1) {
            // Buttons will take care of removing the message
            this.$alertDiv.append(buttons);
        } else {
            // Click on the message anywhere will remove the message
            this.$alertDiv.click(this.removeAlert.bind(this));
        }

        $alertContainer.append(this.$alertDiv);
        this.$alertDiv.fadeIn(this.fadeInDuration, this.afterFadeIn.bind(this));
        this.shown = true;


    };

    /**
     * @private
     */
    n.Alert.prototype.afterFadeIn = function () {
        if (this.timeout > 0) {
            this.timerId = window.setTimeout(this.removeAlert.bind(this), this.timeout);
        }
    };

    /**
     *
     * @param event
     *
     * @private
     */
    n.Alert.prototype.removeAlert = function (event) {
        if (!event || event.type !== "click") {
            // No user click, so it must be a timer event
            if (this.timerId) {
                window.clearTimeout(this.timerId);
                this.timerId = null;
            } else {
                QfqNS.Log.error("Alert.remove(): Identified timer event, but had no timer id");
            }
        }
        var that = this;
        this.$alertDiv.fadeOut(this.fadeOutDuration, function () {
            that.$alertDiv.remove();
            that.$alertDiv = null;
            that.shown = false;

            if (that.countAlertsInAlertContainer() === 0) {
                that.removeAlertContainer();
            }
        });

    };

    n.Alert.prototype.addOkButtonHandler = function (handler) {
        this.userOkButtonHandlers.addFunction(handler);
    };

    n.Alert.prototype.addCancelButtonHandler = function (handler) {
        this.userCancelButtonHandlers.addFunction(handler);
    };

    n.Alert.prototype.addSaveButtonHandler = function (handler) {
        this.userSaveButtonHandlers.addFunction(handler);
    };


    /**
     *
     * @param handler
     *
     * @private
     */
    n.Alert.prototype.okButtonHandler = function (handler) {
        this.removeAlert();
        this.userOkButtonHandlers.call(this);
    };

    /**
     *
     * @param handler
     *
     * @private
     */
    n.Alert.prototype.saveButtonHandler = function (handler) {
        this.removeAlert();
        this.userSaveButtonHandlers.call(this);
    };

    /**
     *
     * @param handler
     *
     * @private
     */
    n.Alert.prototype.cancelButtonHandler = function (handler) {
        this.removeAlert();
        this.userCancelButtonHandlers.call(this);
    };

    n.Alert.prototype.isShown = function () {
        return this.shown;
    };

})(QfqNS);