/**
 * @author Rafael Ostertag <rafael.ostertag@math.uzh.ch>
 */

/* global describe */
/* global it */
/* global expect */
/* global QfqNS */
/* global beforeAll */
/* global beforeEach */
/* global jasmine */
/* global $ */

describe("Form", function () {
    'use strict';

    var form;

    beforeEach(function () {
        form = new QfqNS.Form("myForm");
    });

    it("should throw exception upon unknown id", function () {
        expect(function () {
            var form = new QfqNS.Form("doesnotexist");
        }).toThrowError(Error);
    });

    it("should throw exception upon id of non-form element", function () {
        expect(function () {
            var form = new QfqNS.Form("qfqTabs");
        }).toThrowError(Error);
    });

    it("should have formChanged() === false upon initialization", function () {
        expect(form.getFormChanged()).toBe(false);
    });

    it("should set the proper state upon form state change", function () {
        expect(form.getFormChanged()).toBe(false);
        $('#firstname').change();
        expect(form.getFormChanged()).toBe(true);
    });

    it("should notify upon form change", function () {
        var changeHandler = jasmine.createSpy('ChangeHandler');

        form.addChangeHandler(changeHandler);

        $("#name").change();

        expect(changeHandler).toHaveBeenCalledTimes(1);
        expect(changeHandler).toHaveBeenCalledWith(form);
    });

    it("should notify upon state reset", function () {
        var resetHandler = jasmine.createSpy('StateResetHandler');

        form.addResetHandler(resetHandler);

        form.resetFormChanged();

        expect(resetHandler).toHaveBeenCalledTimes(1);
        expect(resetHandler).toHaveBeenCalledWith(form);
    });

    it("should reset state upon resetFormChanged()", function () {
        $("#name").change();


        expect(form.getFormChanged()).toBe(true);
        form.resetFormChanged();
        expect(form.getFormChanged()).toBe(false);
    });

    it("should call user handler on successful submit", function () {
        jasmine.Ajax.install();

        var successHandler = jasmine.createSpy("success");
        var failureHandler = jasmine.createSpy("failure");

        form.addSubmitSuccessHandler(successHandler);
        form.addSubmitFailureHandler(failureHandler);
        form.submitTo("/test/url");

        jasmine.Ajax.requests.mostRecent().respondWith({
            "status": 200,
            "contentType": "application/json",
            "responseText": "{ \"status\": \"success\" }"
        });

        expect(successHandler).toHaveBeenCalledWith(form, {status: "success"}, "success");
        expect(failureHandler).not.toHaveBeenCalled();

        jasmine.Ajax.uninstall();
    });

    it("should call user handler on failed submit", function () {
        jasmine.Ajax.install();

        var successHandler = jasmine.createSpy("success");
        var failureHandler = jasmine.createSpy("failure");

        form.addSubmitSuccessHandler(successHandler);
        form.addSubmitFailureHandler(failureHandler);
        form.submitTo("/test/url");

        jasmine.Ajax.requests.mostRecent().respondWith({
            "status": "404",
            "statusText": "not found",
            "contentType": "application/json",
            "responseText": "{ \"status\": \"not found\" }"
        });

        expect(successHandler).not.toHaveBeenCalled();
        expect(failureHandler).toHaveBeenCalled();
        expect(failureHandler.calls.argsFor(0)[0]).toBe(form);
        expect(failureHandler.calls.argsFor(0)[1]).toBe("error");
        // arg 2 is jqXHR, which I don't want to test for, and therefore continue with 3
        expect(failureHandler.calls.argsFor(0)[3]).toBe("not found");

        jasmine.Ajax.uninstall();
    });
});