<?php
/**
 * Created by PhpStorm.
 * User: crose
 * Date: 1/12/16
 * Time: 4:36 PM
 */

namespace qfq;

use qfq;
use qfq\Store;

require_once(__DIR__ . '/../qfq/store/Store.php');
require_once(__DIR__ . '/../qfq/Database.php');

/**
 * Class Evaluate
 * @package qfq
 */
class Evaluate {
    private $store = null;
    private $db = null;
    private $startDelimiter = '';
    private $startDelimiterLength = 0;
    private $endDelimiter = '';
    private $endDelimiterLength = 0;
    private $sqlKeywords = array('SELECT ', 'INSERT ', 'DELETE ', 'UPDATE ', 'SHOW ');

//    private $debugStack = array();


    public function __construct(Store $store, Database $db, $startDelimiter = '{{', $endDelimiter = '}}') {
        $this->store = $store;
        $this->db = $db;
        $this->startDelimiter = $startDelimiter;
        $this->startDelimiterLength = strlen($startDelimiter);
        $this->endDelimiter = $endDelimiter;
        $this->endDelimiterLength = strlen($endDelimiter);
    }

    /**
     * Evaluate a whole array or a array of arrays.
     *
     * @param $tokenArray
     * @return mixed
     */
    public function parseArray($tokenArray, &$debugStack = array()) {
        $arr = array();

        foreach ($tokenArray as $key => $value) {
            if (is_array($value)) {
                $arr[] = $this->parseArray($value);
            } else {
                $arr[$key] = $this->parse($value, 0, $debugStack);
            }
        }

        return $arr;
    }

    /**
     * Recursive evaluation of 'line'.
     *
     * Token to replace have to be enclosed by '{{' and '}}'
     *
     * @param $line
     * @param int $recursion
     * @return array|mixed|null|string
     * @throws UserException
     */
    public function parse($line, $recursion = 0, &$debugStack = array()) {
        $flagTokenReplaced = false;

        if ($recursion > 4) {
            throw new qfq\UserException("Recursion too deep: $line", ERROR_RECURSION_TOO_DEEP);
        }

        $result = $line;

        $debugIndent = str_repeat(' ', $recursion);
        $debugLocal[] = $debugIndent . "#Parse: '$result'";

        $posFirstClose = strpos($result, $this->endDelimiter);

        while ($posFirstClose !== false) {
            $flagTokenReplaced = true;

            $posMatchOpen = strrpos(substr($result, 0, $posFirstClose), $this->startDelimiter);
            if ($posMatchOpen === false) {
                throw new \qfq\UserException("Missing open delimiter: $result", ERROR_MISSING_OPEN_DELIMITER);
            }

            $pre = substr($result, 0, $posMatchOpen);
            $post = substr($result, $posFirstClose + $this->endDelimiterLength);
            $match = substr($result, $posMatchOpen + $this->startDelimiterLength, $posFirstClose - $posMatchOpen - $this->startDelimiterLength);

            $evaluated = $this->substitute($match);
            $debugLocal[] = $debugIndent . "#Replace: '$match'";

            // If an array is returned, break everything and return this assoc array.
            if (is_array($evaluated)) {
                $result = $evaluated;
                $debugLocal[] = $debugIndent . "#By: 'array(" . count($result) . ")'";
                break;
            }

            $debugLocal[] = $debugIndent . "#By: '$evaluated'";

            // More to substitute in the new evaluated result? Start recursion just with the new result..
            if (strpos($evaluated, $this->endDelimiter) !== false) {
                $evaluated = $this->parse($evaluated, $recursion + 1, $debugLocal);
            }  

            $result = $pre . $evaluated . $post;
            $posFirstClose = strpos($result, $this->endDelimiter);
        }

        if ($flagTokenReplaced === true) {
            $debugLocal[] = $debugIndent . "#Final: " . is_array($result) ?  "array(" . count($result) .")" : "'$result'" ;
            $debugStack = $debugLocal;
        }

        return $result;
    }

    /**
     * Tries to substitute $token.
     * Token might be
     *   a) fetch from a store. Syntax: 'form', 'form:C', 'form:SC0', 'form:S:ALNUMX'
     *   b) a SQL statement to fire
     * The token have to be _without_ Delimiter '{{' / '}}'
     * If neither a) or b) match, return the token itself, surrounded by single ticks, to emphase that substition failed.
     *
     * @param $token
     * @return array|mixed|null|string
     * @throws DbException
     */
    public function substitute($token) {
        $sqlMode = ROW_IMPLODE_ALL;

        $token = trim($token);

        // just to extract the first token: check if this is a SQL Statement
        $arr = explode(' ', $token, 2);

        if ($token[0] === '!') {
            $token = substr($token, 1);
            $arr[0] = substr($arr[0], 1);
            $sqlMode = ROW_REGULAR;
        }

        // SQL Statement?
        if (in_array(strtoupper($arr[0] . ' '), $this->sqlKeywords)) {
            return $this->db->sql($token, $sqlMode);
        }

        // explode for: <key>:<store priority>:<sanitize class>
        $arr = explode(':', $token, 3);
        if (!isset($arr[1]))
            $arr[1] = null;
        if (!isset($arr[2]))
            $arr[2] = null;


        // search for value in stores
        $value = $this->store->getVar($arr[0], $arr[1], $arr[2]);

        // nothing replaced: put ticks around, to sanitize strings for SQL statements. Nothing to substitute is not a wished situation.
        return ($value === false) ? "'" . $token . "'" : $value;
    }

    public function getDebug() {
        return '<pre>' . implode("\n", $this->debugStack) . '</pre>';
    }
}